'use strict'

let mysql = require('mysql');
let printf = require('printf');
let redis = require('redis');

let g_pool = null;
let g_redisClient = null;

let g_cacheTableMap = {};

function initRedis(host, port) {
    g_redisClient = redis.createClient(port, host);
}

function addCacheTable(table) {
    g_cacheTableMap[table] = 1;
}

function isCacheTable(table) {
    return g_redisClient != null && g_cacheTableMap[table] == 1;
}

function makeCacheKey(qb) {
    return [qb.table, ':', qb.makeWhereStr(), qb.makeGroupStr(), qb.makeOrderStr(), qb.makeLimitStr(), qb.queryFields].join("");
}
function makeCacheDelKey(qb) {
    return qb.table + ':*';
}

function initDb(host, port, user, passwd, database, connectionLimit) {
    g_pool = mysql.createPool({
        host: host,
        port: port,
        user: user,
        password: passwd,
        database: database,
        connectionLimit: connectionLimit,
    });
}

function createQueryExecutor(qb) {
    let qe = new QueryExecutor();
    qe.qb = qb;
    return qe;
}

function QueryExecutor() {
    this.qb = null;
}

QueryExecutor.prototype.p_findOne = function() {
    return this.p_findAll().then(function(queryRes) {
        return new Promise(function(resolve, reject) {
            if (queryRes.length > 0) {
                resolve(queryRes[0]);
            } else {
                resolve(null);
            }
        });
    });
}

QueryExecutor.prototype.p_findAll = function() {
    let _this = this;
    return new Promise(function(resolve, reject) {
        if (isCacheTable(_this.qb.table)) {
            let key = makeCacheKey(_this.qb);
            g_redisClient.get(key, function(err, reply) {
                if (err) {
                    reject(err);
                }
                if (reply != null) {
                    resolve(JSON.parse(reply));
                } else {
                    let str = _this.qb.toQueryStr();
                    g_pool.query(str, function(err, res) {
                        if (err) {
                            reject(err);
                        } else {
                            g_redisClient.set(key, JSON.stringify(res));
                            resolve(res);
                        }
                    });
                }
            });
        } else {
            let str = _this.qb.toQueryStr();
            g_pool.query(str, function(err, res) {
                if (err) {
                    reject(err);
                } else {
                    resolve(res);
                }
            });
        }
    });
};

QueryExecutor.prototype.p_insert = function() {
    let _this = this;
    return new Promise(function(resolve, reject) {
        let str = _this.qb.toInsertStr();
        g_pool.query(str, function(err, res) {
            if (err) {
                reject(err);
            } else {
                if (isCacheTable(_this.qb.table)) {
                    let delKey = makeCacheDelKey(_this.qb);
                    g_redisClient.keys(delKey, function(err, replies) {
                        if (err) {
                            reject(err);
                        } else {
                            for (let i = 0; i < replies.length; i++) {
                                g_redisClient.set(replies[i], null);
                            }
                        }
                    });
                }

                resolve(res);
            }
        });
    });
};

QueryExecutor.prototype.p_update = function() {
    let _this = this;
    return new Promise(function(resolve, reject) {
        let str = _this.qb.toUpdateStr();
        g_pool.query(str, function(err, res) {
            if (err) {
                reject(err);
            } else {
                if (isCacheTable(_this.qb.table)) {
                    let delKey = makeCacheDelKey(_this.qb);
                    g_redisClient.keys(delKey, function(err, replies) {
                        if (err) {
                            reject(err);
                        } else {
                            for (let i = 0; i < replies.length; i++) {
                                g_redisClient.set(replies[i], null);
                            }
                        }
                    });
                }

                resolve(res);
            }
        });
    });
};

QueryExecutor.prototype.p_delete = function() {
    let _this = this;
    return new Promise(function(resolve, reject) {
        let str = _this.qb.toDeleteStr();
        g_pool.query(str, function(err, res) {
            if (err) {
                reject(err);
            } else {
                if (isCacheTable(_this.qb.table)) {
                    let delKey = makeCacheDelKey(_this.qb);
                    g_redisClient.keys(delKey, function(err, replies) {
                        if (err) {
                            reject(err);
                        } else {
                            for (let i = 0; i < replies.length; i++) {
                                g_redisClient.set(replies[i], null);
                            }
                        }
                    });
                }

                resolve(res);
            }
        });
    });
};


function createQueryBuilder(table) {
    let qb = new QueryBuilder();
    qb.table = table;
    return qb;
}


function unserializeQueryBuilder(code) {
    let b = new Buffer(code, 'base64');
    let str = b.toString();
    let obj = JSON.parse(str);
    let qb = new QueryBuilder();
    for (let key in obj) {
        qb[key] = obj[key];
    }
    return qb;
}

function QueryBuilder() {
    this.table = "";
    this.queryFields = "";
    this.queryAdds = [];
    this.querySets = [];
    this.queryOrders = [];
    this.queryGroup = "";
    this.queryLimit = {
        offset: 0,
        count: 0,
    }
    this.queryWheres = [];
    return this;
}

QueryBuilder.prototype.serialize = function(col, value) {
    let str = JSON.stringify(this);
    let b = new Buffer(str)
    return b.toString('base64');
};

QueryBuilder.prototype.add = function(col, value) {
    this.queryAdds.push({
        col: col,
        param: value,
    });
    return this;
};

QueryBuilder.prototype.set = function(col, value) {
    this.querySets.push({
        col: col,
        param: value,
    });
    return this;
};

QueryBuilder.prototype.fields = function(cols) {
    this.queryFields = cols;
    return this;
};

QueryBuilder.prototype.group = function(cols) {
    this.queryGroup = cols;
    return this;
};


QueryBuilder.prototype.order = function(by, direction) {
    this.queryOrders.push({
        by: by,
        direction: direction,
    });
    return this;
};

QueryBuilder.prototype.limit = function(offset, count) {
    this.queryLimit.offset = offset;
    this.queryLimit.count = count;
    return this;
};

QueryBuilder.prototype.nextPage = function() {
    this.queryLimit.offset = offset += this.queryLimit.count;
    return this;
};

QueryBuilder.prototype.where = function(col, op, param) {
    this.queryWheres.push({
        col: col,
        op: op,
        param: param,
    });
    return this;
};

QueryBuilder.prototype.makeWhereStr = function() {
    let whereArr = [];
    for (let i = 0; i < this.queryWheres.length; i++) {
        let qw = this.queryWheres[i];
        let str = "";
        if (typeof(qw.param) == "string") {
            str = printf("`%(col)s` %(op)s '%(param)s'", qw);
        } else {
            str = printf("`%(col)s` %(op)s %(param)s", qw);
        }
        whereArr.push(str);
    }
    let whereStr = "";
    if (whereArr.length > 0) {
    	whereArr.sort();
        whereStr = " WHERE " + whereArr.join(" AND ");
    }
    return whereStr;
};

QueryBuilder.prototype.toInsertStr = function() {
    let addColArr = [];
    let addValArr = [];

    for (let i = 0; i < this.queryAdds.length; i++) {
        let qa = this.queryAdds[i];
        addColArr.push('`' + qa.col + '`');
        if (typeof(qa.param) == "string") {
            addValArr.push(printf("'%(param)s'", qa));
        } else {
            addValArr.push(printf("%(param)s", qa));
        }
    }
    return printf("INSERT INTO `%s` (%s) VALUES(%s)", this.table, addColArr.join(","), addValArr.join(","));
};

QueryBuilder.prototype.toUpdateStr = function() {
    let setArr = [];
    for (let i = 0; i < this.querySets.length; i++) {
        let qs = this.querySets[i];
        if (typeof(qs.param) == "string") {
            setArr.push(printf("`%(col)s`='%(param)s'", qs));
        } else {
            setArr.push(printf("`%(col)s`=%(param)s", qs));
        }
    }
    let setStr = "";
    if (setArr.length > 0) {
        setStr = " SET " + setArr.join(",");
    }

    let whereStr = this.makeWhereStr();
    return printf("UPDATE `%s`%s%s", this.table, setStr, whereStr);
};

QueryBuilder.prototype.toDeleteStr = function() {
    let whereStr = this.makeWhereStr();
    return printf("DELETE FROM `%s`%s", this.table, whereStr);
};

QueryBuilder.prototype.makeOrderStr = function() {
    let orderArr = [];
    for (let i = 0; i < this.queryOrders.length; i++) {
        let qo = this.queryOrders[i];
        orderArr.push(printf("`%s` %s", qo.by, qo.direction));
    }
    let orderStr = "";
    if (orderArr.length > 0) {
    	orderArr.sort();
        orderStr = " ORDER BY " + orderArr.join(",");
    }
    return orderStr;
};

QueryBuilder.prototype.makeLimitStr = function() {
    let limitStr = "";
    if (this.queryLimit.count > 0) {
        limitStr = printf(" LIMIT %d,%d", this.queryLimit.offset, this.queryLimit.count);
    }
    return limitStr;
};

QueryBuilder.prototype.makeGroupStr = function() {
    let groupStr = "";
    if (this.queryGroup != "") {
        groupStr = " GROUP BY `" + this.queryGroup.replace(",", "`,`") + "`";
    }
    return groupStr;
};

QueryBuilder.prototype.toQueryStr = function() {
    let whereStr = this.makeWhereStr();

    let limitStr = this.makeLimitStr();

    let orderStr = this.makeOrderStr();

    let fieldStr = "*";
    if (this.queryFields != "" && this.queryFields != fieldStr) {
        let fieldArr = this.queryFields.split(",");
        for (let i = 0; i < fieldArr.length; i++) {
            let col = fieldArr[i];
            if (col.indexOf('(') > 0 && col.indexOf(')') > 0) {
                continue;
            } else {
                fieldArr[i] = "`" + col + "`";
            }
        }
        fieldStr = fieldArr.join(',');
    }

    let groupStr = this.makeGroupStr();

    return printf("SELECT %s FROM `%s`%s%s%s%s", fieldStr, this.table, whereStr, groupStr, orderStr, limitStr).trim();
};

module.exports = {
    initDb: initDb,
    initRedis: initRedis,
    addCacheTable: addCacheTable,
    createQueryExecutor: createQueryExecutor,
    createQueryBuilder: createQueryBuilder,
    unserializeQueryBuilder: unserializeQueryBuilder,
}